/* @(#)scda.c	15.1.1.1 (ESO-DMD) 01/12/01 17:08:25 */
/*===========================================================================
  Copyright (C) 1995 European Southern Observatory (ESO)
 
  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public 
  License along with this program; if not, write to the Free 
  Software Foundation, Inc., 675 Massachusetts Ave, Cambridge, 
  MA 02139, USA.
 
  Correspondence concerning ESO-MIDAS should be addressed as follows:
	Internet e-mail: midas@eso.org
	Postal address: European Southern Observatory
			Data Management Division 
			Karl-Schwarzschild-Strasse 2
			D 85748 Garching bei Muenchen 
			GERMANY
===========================================================================*/

/*++++++++++++++++++++++++++ SC interface module SCDA +++++++++++++++++++++++
.LANGUAGE C
.IDENTIFICATION Module SCDA
.AUTHOR         K. Banse 		ESO - Garching
.KEYWORDS       standard interfaces, descriptors
.ENVIRONMENT    VMS and UNIX
.COMMENTS
holds SCDCOP, ZSCDCOP, SCDFND, SCDINF, ZSCDINF
.VERSION  [1.00] 920211:  pulled out from scd.c

001002

----------------------------------------------------------------------------*/

#include <fileexts.h>

#include <stdlib.h>


#define DATA_ISIZE   2048			/* 2048 int's  */
#define DATA_CSIZE   DATA_ISIZE*II_SIZE		/* in bytes */
#define DATA_DSIZE   DATA_CSIZE/DD_SIZE		/* no. of double's */

#define DISK_REC     512			/* size of disk record */

static int  mapdat = -1;

static char *fcbsave, *datpntr;

static int  maxdim[4] = {DATA_ISIZE,DATA_ISIZE,DATA_CSIZE-1,DATA_DSIZE};


/*

*/

SCDCOP(from,to,mask,dsc)

/*++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE
Copy descriptors from one frame to another
.ALGORITHM
Depending on 'mask' the relevant descriptors are read from
the source frame + written to the destination frame.
.REMARKS
mask for copying: = 1 , copy all descriptors
                  = 2 , copy only standard desc.
                  = 3 , copy all but standard descriptors
                  = 4 , copy descriptor specified by DSClist
                  = 5 , copy all but extended list of standard descriptors
.RETURNS
return status  ( 0 = o.k. )
--------------------------------------------------*/

int   from;		/* IN: no. of source frame  */
int   to;		/* IN: no. of destination frame  */
int   mask;		/* IN: copy_mask  */
char	*dsc;		/* IN: name of descr. to copy (if mask = 4) */

{
int   iav, lk, nval, nfr, no_elem, byt_elem;
int   total, nin, dblock, dindx, dunit, status, mm, e_c, e_l;
int   nullo;
int   *ibuf;

float *rbuf;

double *dbuf;

char  *cbuf, dtype[4], worky[16], filty, dscbuf[40], descr[16];

static int  def_total[5] = {10000,6,10000,1,10000};

static char  stand0[16] = "NAXIS          ";
static char  stand1[16] = "NPIX           ";
static char  stand2[16] = "START          ";
static char  stand3[16] = "STEP           ";
static char  stand4[16] = "CUNIT          ";
static char  stand5[16] = "IDENT          ";
static char  stand6[16] = "LHCUTS         ";
static char  stand7[16] = "DISPLAY_DATA   ";
static char  stand8[16] = "HIST_BINS      ";
static char  stand9[16] = "HISTOGRAM      ";
static char  stand10[16] = "STATISTIC      ";
static char  stand11[16] = "WINDOW_TO      ";
static char  stand12[16] = "WINDOW_FROM    ";

static char  *stand[13] = {stand0,stand1,stand2,stand3,stand4,stand5,stand6,
                           stand7,stand8,stand9,stand10,stand11,stand12};
static char  tstand0[16] = "TBLENGTH       ";
static char  tstand1[16] = "TBLOFFST       ";
static char  tstand2[16] = "TBLCONTR       ";
static char  tstand3[16] = "TSELTABL       ";

static char  *tstand[4] = {tstand0,tstand1,tstand2,tstand3};

struct FCT_STRUCT  *fctpntr, *fctpntri;

struct FCB_STRUCT  *fcbp;



 

if (from == to) return ERR_NORMAL;		/* this is also possible  */

fctpntri = FCT.ENTRIES + from;
fctpntr = FCT.ENTRIES + to;

if (fctpntr->PROT == 2) return ERR_FILPRO;	/* dest. is write protected */

if (mapdat < 0) 			/* allocate working buffer */
   {					/* for descr. data */
   datpntr = malloc((unsigned int)DATA_CSIZE);
   mapdat = 1;
   }


fcbp = fctpntri->FZP;

if (fcbp->DSCFLAG == 'Z') return ZSCDCOP(from,to,mask,dsc);


if ( (mask < 1 ) || (mask > 5) ) 
   {
   status = ERR_INPINV;
   goto end_of_it;
   }

cbuf = datpntr;
ibuf = (int *) datpntr;
rbuf = (float *) datpntr;
dbuf = (double *) datpntr;


/* disable error abort  */

e_c = ERRO_CONT; e_l = ERRO_LOG;  ERRO_CONT = 1; ERRO_LOG = 0;


/*  determine total no. of descriptors to be copied from `mask' */

if (mask == 4)		/* pad blanks in the end of given descr.  */
   {
   mm = (int) strlen(dsc);
   if (mm > 15) mm = 15;		/* old descr. length = 15 */
   CGN_FILL(worky,' ',15);
   (void) strncpy(worky,dsc,mm);
   worky[15] = '\0';	/* just for completeness... */
   }

total = def_total[mask-1];
filty = fctpntri->CATALOG[0];			/*  get file type */


/*  main loop - read descriptor name + copy, if name o.k.  */

for (lk=1; lk<32767; lk++)
   {
   if (total < 1) break;		/* all descriptors copied...  */


   /* get NAME,TYPE + SIZE back */

   status = SCDINF(from,lk,5,dscbuf,39,ibuf);
   if ( (status != ERR_NORMAL) || (dscbuf[0] == ' ') )
      break;            	/* finished or problem with SCDINF  */

   mm = CGN_INDEXC(dscbuf,',');
   dscbuf[mm] = '\0';
   (void) strcpy(descr,dscbuf);
   dtype[0] = dscbuf[mm+1];
   no_elem = *ibuf;
   byt_elem = *(ibuf+1);


   /*  if 'mask' = 4, compare descriptor with 'dsc'  */

   if (mask == 4) 
      {
      if (strncmp(worky,descr,15) != 0) goto sect_1000;

      total = 0;
      }


   /* if mask = 3, compare descriptor with list of standard descriptors  */

   else if (mask == 3) 
      {
      if (filty == 'T')
         {
         for (nin=0; nin<4; nin++)
            {
            if ((strncmp(descr,tstand[nin],15) == 0)  || 
                (strncmp(descr,"TLABL",5) == 0)) goto sect_1000;
            }
         }
      else
         {
         for (nin=0; nin<7; nin++)
            {
            if (strncmp(descr,stand[nin],15) == 0) goto sect_1000;
            }
         } 				/* non-standard descriptor found  */
      }


   /* if mask = 5, compare descriptor with extended list of stand. descr.  */

   else if (mask == 5) 
      {
      for (nin=0; nin<13; nin++)
         {
         if (strncmp(descr,stand[nin],15) == 0) goto sect_1000;
         }				/* non-standard descriptor found  */
      }


   /*  if mask = 2, compare descriptor with list of standard descriptors  */

   else if (mask == 2)
      {
      if (filty == 'T')
         {
         for (nin=0; nin<4; nin++)
            {
            if (strncmp(descr,tstand[nin],15) == 0)
               {
               total --;
               goto sect_200;
               }
            }
         }
      else
         {
         for (nin=0; nin<7; nin++)
            {
            if (strncmp(descr,stand[nin],15) == 0)
               {
               total --;
               goto sect_200;
               }
            }
         }
      goto sect_1000;
      }


   /*  mask = 1, all source descriptors are copied  */

sect_200:
   nfr = 1;


sect_222:
   nval = no_elem - nfr + 1;
   nullo = -1;
   if (dtype[0] == 'I') 		   /*  integer descriptor  */
      {
      if (nval > maxdim[0]) nval = maxdim[0];

      status = SCDRDI(from,descr,nfr,nval,&iav,ibuf,&dunit,&nullo);
      if (status == ERR_NORMAL) 
         status = SCDWRI(to,descr,ibuf,nfr,nval,&dunit);
      }

   else if (dtype[0] == 'R') 		/* real descriptor  */
      {
      if (nval > maxdim[1]) nval = maxdim[1];

      status = SCDRDR(from,descr,nfr,nval,&iav,rbuf,&dunit,&nullo);
      if (status == ERR_NORMAL) 
         status = SCDWRR(to,descr,rbuf,nfr,nval,&dunit);
      }

   else if (dtype[0] == 'C')   		/*   character descriptor  */
      {
      if (byt_elem > 1)
         {
         mm = maxdim[2] / byt_elem;
         if (mm < 0)
            {
            (void) sprintf
            (cbuf,"bytes_per_element = %d > 4096, descr %s not copied!",
             byt_elem,descr);
            SCTPUT(cbuf);
            goto help_descr;
            }
         }
      else
         mm = maxdim[2];

      if (nval > mm) nval = mm;

      status = SCDRDC(from,descr,
                      byt_elem,nfr,nval,&iav,cbuf,&dunit,&nullo);
      if (status == ERR_NORMAL) 
         {
         cbuf[nval*byt_elem] = '\0';
         status = SCDWRC(to,descr,byt_elem,cbuf,nfr,nval,&dunit);
         }
      }

   else					/*   double precision descriptor  */
      {
      if (nval > maxdim[3]) nval = maxdim[3];

      status = SCDRDD(from,descr,nfr,nval,&iav,dbuf,&dunit,&nullo);
      if (status == ERR_NORMAL) 
         status = SCDWRD(to,descr,dbuf,nfr,nval,&dunit);
      }

   if (status != ERR_NORMAL) goto end_of_it;

   
   /*  test, if descriptor copied completely  */

   if ((nval+nfr) <= no_elem) 
      {
      nfr += nval;
      goto sect_222;
      }

help_descr:
   nval = maxdim[2];
   status = SCDRDH(from,descr,1,nval,&iav,cbuf,&mm);
   if (status != ERR_NORMAL) goto end_of_it;
   if (mm > 0)		/* Yes. So copy also the help */
      {
      cbuf[iav] = '\0';
      status = SCDWRH(to,descr,cbuf,1,iav);
      if (status != ERR_NORMAL) goto end_of_it;
      }

sect_1000:
   ;
   }

	
/*  end of main loop  */

end_of_it:
ERRO_CONT = e_c;  ERRO_LOG = e_l;		/* reset error flags  */

if (status != ERR_NORMAL) MID_E2(8,from,descr,status,1);
return status;
}

/*

*/

ZSCDCOP(from,to,mask,dsc)
/*++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE
Copy descriptors from one frame to another
.ALGORITHM
Depending on 'mask' the relevant descriptors are read from
the source frame + written to the destination frame.
.REMARKS
mask for copying: = 1 , copy all descriptors
		  = 2 , copy only standard desc.
                  = 3 , copy all but standard descriptors
		  = 4 , copy descriptors specified by DSClist (max. 8) 
		  = 5 , copy all but extended list of standard descriptors

                  = 6 , copy all descriptors except descr specified by DSClist
                  = 7 , copy standard descriptors, except DSClist
                  = 8 , copy all but standard descriptors, except DSClist
                  = 10, copy all but extended list of standard descriptors,
                        except descr from.DSClist
		  = 100 + mask as described above indicates, that we want
		        to clear existing descr's completely before being
			overwritten with (shorter) source descr's
.RETURNS
return status  ( 0 = o.k. )
--------------------------------------------------*/

int   from;		/* IN: no. of source frame  */
int   to;		/* IN: no. of destination frame  */
int   mask;		/* IN: copy_mask  */
char  *dsc;		/* IN: DSClist = names of descr. to copy 
			       separated by comma */

{
int   iav, nval, nfr, no_elem, byt_elem, cleanem, ncleanem;
int   total, dblock, dindx, hblock, hindx, dunit, status, mm, e_c, e_l;
int   chanl, nullo, stuff_read, dsc_count, hnc, hno_elem, dsc_exist;
int   tcount, ttcount, diroff, dirfirst, dirlen, dirbytes;
int   fcbDFILLED, fcbPTRLDB, fcbDIREXT, dirb2, dirb3;
int   ignore_help, patflag, patlen, pat1flag, pat1len, *ibuf;
register int  nr;
static int  mapdscdir = -1;

float *rbuf;
double *dbuf;

char  *cbuf, dtype[4], filty, descr[88];
char  *zdscpntr, *dscptr[8], *dscbptr;
register char  *cpt, *dpt;
static char  *dscdir;

static char  stand0[] = "NAXIS";
static char  stand1[] = "NPIX";
static char  stand2[] = "START";
static char  stand3[] = "STEP";
static char  stand4[] = "CUNIT";
static char  stand5[] = "IDENT";
static char  stand6[] = "LHCUTS";
static char  stand7[] = "DISPLAY_DATA";
static char  stand8[] = "HIST_BINS";
static char  stand9[] = "HISTOGRAM";
static char  stand10[] = "STATISTIC";
static char  stand11[] = "WINDOW_TO";
static char  stand12[] = "WINDOW_FROM";

static char  *stand[13] = {stand0,stand1,stand2,stand3,stand4,stand5,stand6,
                           stand7,stand8,stand9,stand10,stand11,stand12};
static char  tstand0[] = "TBLENGTH";
static char  tstand1[] = "TBLOFFST";
static char  tstand2[] = "TBLCONTR";
static char  tstand3[] = "TSELTABL";

static char  *tstand[4] = {tstand0,tstand1,tstand2,tstand3};

struct FCT_STRUCT  *fctfrom, *fctto;

struct FCB_STRUCT  *fcbp;
 



 
cbuf = datpntr;
ibuf = (int *) datpntr;
rbuf = (float *) datpntr;
dbuf = (double *) datpntr;

fctfrom = FCT.ENTRIES + from;
if (fctfrom->LINK[0] >= 2)                      /* use father frame */
   {
   from = fctfrom->LINK[1];
   fctfrom = FCT.ENTRIES + from;
   }

filty = fctfrom->CATALOG[0];			/* get file type */
chanl = fctfrom->IOCHAN;
fcbp = fctfrom->FZP;

fctto = FCT.ENTRIES + to;
if (fctto->LINK[0] >= 2)			/* use father frame */
   {
   to = fctto->LINK[1];
   fctto = FCT.ENTRIES + to;
   }
if (from == to) return ERR_NORMAL;		/* frames are equal */


status = ERR_INPINV;
e_c = ERRO_CONT;   ERRO_CONT = 1;
e_l = ERRO_LOG;    ERRO_LOG = 0;

total = 1; 
patflag = 0;
dsc_exist = 1;				/* default => SCDWRz */
if (mask > 99)
   {
   cleanem = 1;
   mask -= 100;
   }
else
   cleanem = 0;

if ((KEYALL.ORIGIN != -1) && (KIWORDS[OFF_MONPAR+13] == 1))
   ignore_help = 1;
else
   ignore_help = 0;


if (mask > 5)
   {
   tcount = (int) strlen(dsc);
   mask -= 5;
   }
else
   tcount = 0; 


switch (mask)			/* sort out mask business, */
   {				/* set tcount as necessary */
   case 1:
   case 3:
   case 5:
    break;

   case 2:
    if (filty == 'T')
       total = 4;
    else
       total = 7;
    break;

   case 4:				/* pass DSClist */
    tcount = (int) strlen(dsc);
    if (tcount < 1) 			/* we need a descr name */
       goto end0;
    break;

   default:
    mask = 1;
   }


if (tcount > 0)				/* we got descriptors in DSClist */
   {		
   for (nr=0; nr<tcount; nr++)
      {
      if (dsc[nr] == ' ')	/* cut off trailing blanks (FORTRAN...) */
         {
         tcount = nr;
         if (tcount < 1) goto end_of_it;
         break;
         }
      }

   mm = CGN_INDEXC(dsc,'|');		/* do we have incl-patt|excl-patt? */
   if (mm > 1)				/* only considered for mask=4 ...  */
      {
      dscptr[0] = malloc((unsigned int)tcount);
      dscptr[1] = malloc((unsigned int)tcount);

      if (dsc[0] == '*')
         {
         if (dsc[mm-1] == '*')		/* we have 1. pattern  `*abc*'  */
            {
            patflag = 11;
            patlen = mm - 2;
            }
         else				/* we have 1. pattern  `*abc '  */
            {
            patflag = 12;
            patlen = mm - 1;
            }
         if (patlen < 1) goto patt_end;

         CGN_UPCOPY(dscptr[0],dsc+1,patlen);  /* copy in upper case */
         *(dscptr[0]+patlen) = '\0';
         }
      else if (dsc[mm-1] == '*')	/* we have 1. pattern  ` abc*'  */
         {
         patflag = 13;
         patlen = mm - 1;
         if (patlen < 1) goto patt_end;
      
         CGN_UPCOPY(dscptr[0],dsc,patlen);
         *(dscptr[0]+patlen) = '\0';
         }

      mm ++;					/* point to excl-pattern */
      if (dsc[mm] == '*')
         {
         if (dsc[tcount-1] == '*')	/* we have 2. pattern  `*abc*'  */
            {
            pat1flag = 1;
            pat1len = tcount - mm - 2;
            }
         else				/* we have 2. pattern  `*abc '  */
            {
            pat1flag = 2;
            pat1len = tcount - mm - 1;
            }
         if (pat1len < 1) goto patt_end;

         CGN_UPCOPY(dscptr[1],dsc+mm+1,pat1len);
         *(dscptr[1]+pat1len) = '\0';
         }
      else if (dsc[tcount-1] == '*')	/* we have 2. pattern  ` abc*'  */
         {
         pat1flag = 3;
         pat1len = tcount - mm - 1;
         if (pat1len < 1) goto patt_end;
      
         CGN_UPCOPY(dscptr[1],dsc,pat1len);
         *(dscptr[1]+pat1len) = '\0';
         }

      tcount = 2;			/* we use two pointers */
      goto do_FCB_stuff;

     patt_end:				/* problems, problems, ...  */
      free (dscptr[0]);
      free (dscptr[1]);
      goto end0;
      }
 

   if (dsc[0] == '*')
      {
      if (dsc[tcount-1] == '*')		/* we have a pattern  `*abc*'  */
         {
         patflag = 1;
         patlen = tcount - 2;
         }
      else				/* we have a pattern  `*abc '  */
         {
         patflag = 2;
         patlen = tcount - 1;
         }
      if (patlen < 1) goto end_of_it;
      dscptr[0] = malloc((unsigned int)(patlen+1));
      CGN_UPCOPY(dscptr[0],dsc+1,patlen);  /* copy in upper case to 1. ptr */
      *(dscptr[0]+patlen) = '\0';
      tcount = 1;
      }

   else if (dsc[tcount-1] == '*')	/* we have a pattern  ` abc*'  */
      {
      patflag = 3;
      patlen = tcount - 1;
      if (patlen < 1) goto end_of_it;
      dscptr[0] = malloc((unsigned int)(patlen+1));
      CGN_UPCOPY(dscptr[0],dsc,patlen);	/* copy in upper case to 1. ptr */
      *(dscptr[0]+patlen) = '\0';
      tcount = 1;
      }

   else
      {					/* list of up to 8 descrs' */
      dscptr[0] = malloc((unsigned int)(tcount+1));
      CGN_UPCOPY(dscptr[0],dsc,tcount);	/* copy in upper case to 1. ptr */
      *(dscptr[0]+tcount) = '\0';

      cpt = dscptr[0];
      tcount = 1;			/* use as counter of dscptr[] */
      for (nr=1; nr<8; nr++)
         {
         mm = CGN_INDEXC(cpt,',');
         if (mm > 0)
            {
            *(dscptr[nr-1]+mm) = '\0';
            tcount++;
            cpt += (mm+1);
            dscptr[nr] = malloc((unsigned int)80);
            CGN_UPCOPY(dscptr[nr],cpt,80);
            *(dscptr[nr]+79) = '\0';
            }
         }
      if (mask == 4) total = tcount;
      }
   }



do_FCB_stuff:
status = ERR_NORMAL;
fcbDFILLED = fcbp->DFILLED;
if (fcbp->DFILLED > fcbp->DIREXT)
   fcbDIREXT = fcbp->DFILLED;
else
   fcbDIREXT = fcbp->DIREXT;
fcbPTRLDB = fcbp->PTRLDB;
dirb2 = 2 * fcbp->DIRENTRY;
dirb3 = 3 * fcbp->DIRENTRY;


/* allocate working buffer for descr. directory */

if (mapdscdir < 0)                      /* allocate working buffer */
   {                                    /* for descr. directory */
   dscdir = malloc((unsigned int)fcbDIREXT);		/* at least 1500 */
   mapdscdir = 1;
   }
zdscpntr = (char *) &ZDSCDIR_ENTRY;


/*  main loop - read descriptor name + copy, if name o.k.  */


if (mask == 5)
   {
   register int  mr;

   (void) ZSCDINF(fctto->IOCHAN,to,tcount,99,descr,10,&dsc_count);
   dtype[3] = '\0';
   status = SCDRDX(to,1,descr,dtype,&byt_elem,&no_elem,&hnc);
   dscbptr = malloc((unsigned int) (dsc_count*50));	/* holds all descrs */
   dpt = dscbptr;				    /* in destination frame */
   nval = dsc_count;

   for (mr=0; mr<nval; mr++)
      {
      status = SCDRDX(to,10,dpt,dtype,&byt_elem,&no_elem,&hnc);

      if (filty == 'T')			/* compare with descriptors which */
         {				/* are not copied */
         for (nr=0; nr<4; nr++)
            {
            if ((strcmp(dpt,tstand[nr]) == 0) || (strncmp(dpt,"TLABL",5) == 0))
               {
               dsc_count --;
               dpt -= 50;		/* overwrite dsc name again */
               break;
               }
            }
         }
      else
         {
         for (nr=0; nr<13; nr++)
            {
            if (strcmp(dpt,stand[nr]) == 0) 
               {
               dsc_count --;
               dpt -= 50;		/* overwrite dsc name again */
               break;
               }
            }	
         }

      dpt += 50;			/* for next dsc name */
      }

   status = SCDRDX(to,0,descr,dtype,&byt_elem,&no_elem,&hnc);	/* free space */
   }



diroff = 0;
nullo = -1;
dirlen = fcbDFILLED;

(void) MID_RDSCRC(chanl,fcbPTRLDB,0,1,dirlen,dscdir,&nullo);
cpt = dscdir;

if (*cpt == '2')		/* first read gives us descr. directory */
   dirbytes = dirb2;
else 
   dirbytes = dirb3;
cpt += dirbytes;  		/* skip descr. directory */
stuff_read = dirbytes;
dtype[1] = dtype[2] = ' '; dtype[3] = '\0';

read_next_entry:
if (stuff_read < dirlen)
   {
   if (*cpt == '2')
      dirbytes = dirb2;
   else if (*cpt == '3')
      dirbytes = dirb3;
   else 
      dirbytes = fcbp->DIRENTRY;

   stuff_read += dirbytes;
   if (*(cpt+1) != '_')              /* test if just fill_in block */
      {
      CGN_COPYALL(zdscpntr,cpt,dirbytes);	/* get directory entry */
      cpt += dirbytes; 
      (void) strcpy(descr,ZDSC_PNTR->NAME);	/* exactly as in ZDSCDIR */
      dtype[0] = ZDSC_PNTR->TYPE;		/* FIND section */
      byt_elem = ZDSC_PNTR->BYTELEM;
      hno_elem = ZDSC_PNTR->HNOELEM;
      no_elem = ZDSC_PNTR->NOELEM;
      dunit = ZDSC_PNTR->UNIT;
      dblock = ZDSC_PNTR->START;
      dindx = (int) ZDSC_PNTR->INDEX -1;
      hblock = ZDSC_PNTR->HSTART;
      hindx = (int) ZDSC_PNTR->HINDEX -1;
      goto check_mask;
      }
   else
      {
      cpt += dirbytes; 
      goto read_next_entry;
      }
   }

goto end_of_it;					/* we're finished */
 



/*  if 'mask' = 4, compare descriptor with dscptr[0],[1],...  */

check_mask:
if (mask == 4) 
   {
   if (patflag != 0)		/* test for pattern match */
      {
      if (patflag == 3)
         {	
         if (strncmp(descr,dscptr[0],patlen) == 0) goto find_descr;
         }
      else if (patflag == 2)
         {
         mm = (int) strlen(descr) - patlen;
         if (mm > -1)
            {
            if (strcmp(descr+mm,dscptr[0]) == 0) goto find_descr;
            }
         }
      else if (patflag == 1)
         {
         if (CGN_INDEXS(descr,dscptr[0]) > -1) goto find_descr;
         }

      else
         {				/* patflag 11, 12, 13 */
         if (patflag == 13)
            {
            if (strncmp(descr,dscptr[0],patlen) == 0) goto test_excl_pat;
            }
         else if (patflag == 12)
            {
            mm = (int) strlen(descr) - patlen;
            if (mm > -1)
               {
               if (strcmp(descr+mm,dscptr[0]) == 0) goto test_excl_pat;
               }
            }
         else 
            {
            if (CGN_INDEXS(descr,dscptr[0]) > -1) goto test_excl_pat;
            }
         goto read_next_entry;		/* descr not in incl-pattern */
    
        test_excl_pat:
         if (pat1flag == 3)
            {
            if (strncmp(descr,dscptr[1],pat1len) == 0) goto read_next_entry;
            }
         else if (pat1flag == 2)
            {
            mm = (int) strlen(descr) - pat1len;
            if (mm > -1)
               {
               if (strcmp(descr+mm,dscptr[1]) == 0) goto read_next_entry;
               }
            }
         else 
            {
            if (CGN_INDEXS(descr,dscptr[1]) > -1) goto read_next_entry;
            }
         goto find_descr;		/* descr not in excl-pattern */
         }
      }

   else
      {
      for (nr=0; nr<tcount; nr++)
         {
         if (strcmp(dscptr[nr],descr) == 0)
            {
            total --;
            goto find_descr;
            }
         }
      }
   goto read_next_entry;
   }


/*  for all other options check for excluded descrs' */

if (patflag != 0)		/* test for pattern match */
   {
   if (patflag == 3)
      {	
      if (strncmp(descr,dscptr[0],patlen) == 0) goto read_next_entry;
      }
   else if (patflag == 2)
      {
      mm = (int) strlen(descr) - patlen;
      if (mm > -1)
         {
         if (strcmp(descr+mm,dscptr[0]) == 0) goto find_descr;
         }
      }
   else
      {
      if (CGN_INDEXS(descr,dscptr[0]) > -1) goto read_next_entry;
      }
   }

else
   {
   for (nr=0; nr<tcount; nr++)
      {
      if (strcmp(dscptr[nr],descr) == 0) goto read_next_entry;
      }
   }


/*  if mask = 2, compare descriptor with list of standard descriptors
    and only copy these  */

if (mask == 2)
   {
   if (filty == 'T')
      {
      for (nr=0; nr<4; nr++)
         {
         if (strcmp(descr,tstand[nr]) == 0)
            {
            total --;
            goto find_descr;
            }
         }
      }
   else
      {
      for (nr=0; nr<7; nr++)
         {
         if (strcmp(descr,stand[nr]) == 0)
            {
            total --;
            goto find_descr;
            }
         }
      }
   goto read_next_entry;
   }


/* mask = 3, compare descriptor with list of standard descriptors 
   to exclude them from copying */

if (mask == 3) 
   {
   if (filty == 'T')
      {
      for (nr=0; nr<4; nr++)
         {
         if ((strcmp(descr,tstand[nr]) == 0)  || 
             (strncmp(descr,"TLABL",5) == 0)) goto read_next_entry;
         }
      }
   else
      {
      for (nr=0; nr<7; nr++)
         {
         if (strcmp(descr,stand[nr]) == 0) goto read_next_entry;
         }
      }
   }


/* if mask = 5, exclude descriptors of extended list of stand. descrs'
   from copying, and copy all others  */

if (mask == 5) 
   {
   if (filty == 'T')
      {
      for (nr=0; nr<4; nr++)
         {
         if ((strcmp(descr,tstand[nr]) == 0)  || 
             (strncmp(descr,"TLABL",5) == 0)) goto read_next_entry;
         }
      }
   else
      {
      for (nr=0; nr<13; nr++)
         {
         if (strcmp(descr,stand[nr]) == 0) goto read_next_entry;
         }	
      }

   dsc_exist = 0;			/* check, if same dest.descr. exists */
   dpt = dscbptr;			/* start of dest. descr list */

   for (nr=0; nr<dsc_count; nr++)
      {
      if (strcmp(descr,dpt) == 0) 
         {
         dsc_exist = 1;
         break;
         }
      dpt += 50;
      }	
   }



find_descr:
nfr = 1;
ncleanem = cleanem;


data_loop:
nval = no_elem - nfr + 1;
nullo = -1;

if (dtype[0] == 'I')			/*  integer descriptor  */
   {
   if (nval > maxdim[0]) nval = maxdim[0];

   (void) MID_RDSCRI(chanl,dblock,dindx,nfr,nval,ibuf,&nullo);
   if (dsc_exist == 1)
      {
      if (ncleanem == 1) 		/* see, if dest. descr. bigger */
         {
         int  bby, nno, ddu, ddb, ddi, ii;

         status = MID_ZDSCDIR(to,'F',descr,dtype,&bby,&nno,&ddu,&ddb,&ddi);
         ddu = 0;
         if ((status == ERR_NORMAL) && (nno > nval))
            {
            mm = nno - nval;
            ii = 0;
            (void) MID_WDSCRI(fctto->IOCHAN,ddb,ddi,&ii,1,nval+1,mm);
            }
         }
      status = SCDWRI(to,descr,ibuf,nfr,nval,&dunit);
      }
   else
      status = xSCDWRI(to,descr,ibuf,nfr,nval,&dunit);
   }

else if (dtype[0] == 'R') 		/* real descriptor  */
   {
   if (nval > maxdim[1]) nval = maxdim[1];

   (void) MID_RDSCRR(chanl,dblock,dindx,nfr,nval,rbuf,&nullo);
   if (dsc_exist == 1)
      {
      if (ncleanem == 1) 		/* see, if dest. descr. bigger */
         {
         int  bby, nno, ddu, ddb, ddi;
         float  rr;

         status = MID_ZDSCDIR(to,'F',descr,dtype,&bby,&nno,&ddu,&ddb,&ddi);
         ddu = 0;
         if ((status == ERR_NORMAL) && (nno > nval))
            {
            mm = nno - nval;
            rr = 0.;
            (void) MID_WDSCRR(fctto->IOCHAN,ddb,ddi,&rr,1,nval+1,mm);
            }
         }
      status = SCDWRR(to,descr,rbuf,nfr,nval,&dunit);
      }
   else
      status = xSCDWRR(to,descr,rbuf,nfr,nval,&dunit);
   }

else if (dtype[0] == 'C')   		/*   character descriptor  */
   {
   if (byt_elem > 1)
      {
      mm = maxdim[2] / byt_elem;
      if (mm < 0)
         {
         (void) sprintf
         (cbuf,"bytes_per_element = %d > %d, descr %s not copied!",
          byt_elem,maxdim[2],descr);
         SCTPUT(cbuf);
         if (total > 0)
            goto read_next_entry;
         else
            goto end_of_it;
         }
      if (nval > mm) nval = mm;
      mm = nval*byt_elem;
      iav = (nfr-1)*byt_elem + 1;
      }
   else
      {
      if (nval > maxdim[2]) nval = maxdim[2];
      mm = nval;
      iav = nfr;
      }

   
   (void) MID_RDSCRC(chanl,dblock,dindx,iav,mm,cbuf,&nullo);
   cbuf[mm] = '\0';
   if (dsc_exist == 1)
      {
      if (ncleanem == 1) 		/* see, if dest. descr. bigger */
         {
         int  bby, nno, ddu, ddb, ddi;
         char  cc;

         status = MID_ZDSCDIR(to,'F',descr,dtype,&bby,&nno,&ddu,&ddb,&ddi);
         ddu = 0;
         mm = bby * nno;
         if ((status == ERR_NORMAL) && (mm > (byt_elem*nval)))
            {
            bby = byt_elem * nval;
            mm -= bby;
            cc = ' ';
            (void) MID_WDSCRC(fctto->IOCHAN,ddb,ddi,&cc,1,bby+1,mm);
            }
         }
      status = SCDWRC(to,descr,byt_elem,cbuf,nfr,nval,&dunit);
      }
   else
      status = xSCDWRC(to,descr,byt_elem,cbuf,nfr,nval,&dunit);
   }

else					/*   double precision descriptor  */
   {
   if (nval > maxdim[3]) nval = maxdim[3];

   nr = DD_SIZE / II_SIZE;		/* we use integers for double descr */
   iav = (nfr-1)*nr + 1;		/* first element in double buf */
   mm = nval * nr;
   (void) MID_RDSCRI(chanl,dblock,dindx,iav,mm,ibuf,&nullo);

   if (dsc_exist == 1)
      {
      if (ncleanem == 1) 		/* see, if dest. descr. bigger */
         {
         int  bby, nno, ddu, ddb, ddi, ii;

         status = MID_ZDSCDIR(to,'F',descr,dtype,&bby,&nno,&ddu,&ddb,&ddi);
         ddu = 0;
         if ((status == ERR_NORMAL) && (nno > nval))
            {
            mm = (nno - nval) * nr;
            bby = ((nval+1)*nr) - 1;
            ii = 0;
            (void) MID_WDSCRI(fctto->IOCHAN,ddb,ddi,&ii,1,bby,mm);
            }
         }
      status = SCDWRD(to,descr,dbuf,nfr,nval,&dunit);	/* ibuf <-> dbuf */
      }
   else
      status = xSCDWRD(to,descr,dbuf,nfr,nval,&dunit);	/* ibuf <-> dbuf */
   }

if (status != ERR_NORMAL) goto end_of_it;

   
if ((nval+nfr) <= no_elem)	/* test, if descriptor was copied completely */
   {
   nfr += nval;
   dsc_exist = 1;
   ncleanem = 0;
   goto data_loop;
   }


/* finally, look if help text exists for the descriptor we just copied */

if (ignore_help == 0)
   {
   /* if ((hno_elem > 0) && (dsc_exist == 0)) */
   if (hno_elem > 0) 
      {
      nullo = -1;
      (void) MID_RDSCRC(chanl,hblock,hindx,1,hno_elem,cbuf,&nullo);
      cbuf[hno_elem] = '\0';
      status = SCDWRH(to,descr,cbuf,1,hno_elem);
      if (status != ERR_NORMAL) goto end_of_it;
      }
   else
      status = ERR_NORMAL;
   }

if (total > 0) goto read_next_entry;

	

end_of_it:
if (mask == 5) free(dscbptr);
free(dscdir);
mapdscdir = -1;


end0:
ERRO_CONT = e_c;  ERRO_LOG = e_l;		/* reset error flags  */

if (tcount != 0)			/* free the `dscptr' allocations */
   {
   free(dscptr[0]);
   tcount --;
   nr = 1;
   while (tcount > 0)
      {
      free(dscptr[nr++]);
      tcount --;
      }
   }


if (status != ERR_NORMAL) MID_E2(8,from,descr,status,1);
return status;
}

/*

*/

SCDFND(imno,descr,type,noelem,bytelem)

/*++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE
provide info about descriptor 'descr'
.ALGORITHM
use MID_DSCDIR to find descr. + related info
.RETURNS
return status  ( 0 = o.k. )
--------------------------------------------------*/

int   imno	/* IN: no. of data frame  */;
char   *descr   /* IN:	descriptor name  */;
char   *type	/* OUT: type of descr - I, R, C, D or ' ' on output */;
int   *noelem	/* OUT: no. of elements  */;
int   *bytelem	/* OUT: no. of bytes per element  */;

{ 	
int   dblock, dindx, snoel, sbytel;
int   dunit, status, e_c, e_l;

struct FCB_STRUCT  *fcbp;

struct FCT_STRUCT  *fctpntr;



	
/* disable error abort  */

e_c = ERRO_CONT; e_l = ERRO_LOG;  ERRO_CONT = 1; ERRO_LOG = 0;


/*  check imno + type  */
 
if ( (imno < 0) || (imno >= FCT.MAXENT) ) 
   {
   status = ERR_INPINV;
   goto end_of_it;
   }

fctpntr = FCT.ENTRIES + imno;
if (fctpntr->LINK[0] >= 2)			/* use father frame */
   {
   imno = fctpntr->LINK[1];
   fctpntr = FCT.ENTRIES + imno;
   }

*type = ' ';
fcbp = fctpntr->FZP;
if (fcbp->DSCFLAG == 'Z')
   status = MID_ZDSCDIR(imno,'F',descr,type,&sbytel,&snoel,&dunit,
                       &dblock,&dindx);
else
   status = MID_DSCDIR(imno,'F',descr,type,&sbytel,&snoel,&dunit,
                       &dblock,&dindx);

if (status != ERR_DSCNPR)
   {
   *noelem = snoel;
   *bytelem = sbytel;
   }

ERRO_CONT = e_c;  ERRO_LOG = e_l;		/* reset error flags  */
return ERR_NORMAL;	
 
end_of_it:
ERRO_CONT = e_c;  ERRO_LOG = e_l;		/* reset error flags  */
MID_E2(7,imno,"SCDFND: ",status,1);
return status;
}

/*

*/

SCDINF(imno,npos,fno,buf,lbuf,numbuf)

/*++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE
provide info about descriptor at position 'npos'
.ALGORITHM
go through the descriptor directory and provide the desired info
.RETURNS
return status  ( 0 = o.k. )
--------------------------------------------------*/

int  imno	/* IN : no. of data frame  */;
int  npos	/* IN : position of descriptor, beginning with 1 
			use 0 for the descr. directory itself    */;
int  fno	/* IN : specify desired info, 
                        1 = NAME, 2 = TYPE, 3 = SIZE  
                        4 = NAME,TYPE separated by a comma in `buf'
			    and SIZE in `numbuf'
                        5 = NAME,TYPE separated by a comma in `buf'
			    and SIZE,BYTELEM in `numbuf'
		       98 = return total no. of descriptors (except Help-d.)
		       99 = return total no. of descriptors */;
char   *buf	/* IN : buffer for character descriptor info */;
int lbuf	/* IN : max. length of buffer above  */;
int *numbuf	/* IN : return buffer for numerical data  */;

{ 	
int   status, chanl, ext, flag;
int   dirfirst, dirused, dirsize, dirlen, diroff, totext;
int   nullo;
int   snpos;
register int  nr;
static int  mapdscdir = -1;

char  *dscpntr;
register char  *cpt;
static char  *dscdir;

struct FCT_STRUCT  *fctpntr;

struct FCB_STRUCT  *fcbp;




if ( (imno < 0) || (imno >= FCT.MAXENT) ) 
   {
   status= ERR_INPINV;			/* wrong argument `npos'  */
   goto end_of_it;
   }

snpos = npos + 1;	/* real index is at NPOS+1 because of directory */
if ( (snpos < 1) || (fno < 0) )
   {
   status= ERR_INPINV;			/* wrong argument `npos'  */
   goto end_of_it;
   }

status = ERR_NORMAL;
 
fctpntr = FCT.ENTRIES + imno;
if (fctpntr->LINK[0] >= 2)                      /* use father frame */
   {
   imno = fctpntr->LINK[1];
   fctpntr = FCT.ENTRIES + imno;
   }

chanl = fctpntr->IOCHAN;
fcbp = fctpntr->FZP;


if (fcbp->DSCFLAG == 'Z')			/* new descriptor format? */
   return ZSCDINF(chanl,imno,snpos,fno,buf,lbuf,numbuf);

if (fcbp->DSCFLAG == 'N')
   {
   dirused = fcbp->DFILLED;
   dirsize = fcbp->DSIZE;
   }
else
   {
   dirused = fcbp->DIRUSE;
   dirsize = fcbp->DIRNOELEM;
   }

totext = dirsize/fcbp->DIREXT;      /* total extensions for directory */
dscpntr = (char *) &DSCDIR_ENTRY.NAME[0];

ext = 1;
npos = 0;
diroff = 0;
flag = 0;

pos_loop:
dirfirst = diroff + 1;
dirlen = dirused - diroff;
if (fcbp->DIREXT < dirlen) 
   dirlen = fcbp->DIREXT;    /* read max. 1500 chars.  */

if (mapdscdir < 0)                      /* allocate working buffer */
   {                                    /* for descr. directory */
   dscdir = malloc((unsigned int) dirlen);  
   mapdscdir = 1;
   }

if (fno > 90)
   {
   nullo = -1;
   (void) MID_RDSCRC(chanl,fcbp->PTRLDB,0,dirfirst,dirlen,dscdir,&nullo);
   cpt = dscdir;

   for (nr=0; nr<dirlen; nr+=fcbp->DIRENTRY)   /* take chunks of 30 ch. */
      {
      if (*cpt != ' ') 
         {
         if (fno == 99)
            npos ++;
         else
            {
            if (*(cpt+15) != 'H') npos ++;
            }
         }
      cpt += fcbp->DIRENTRY;
      }


   /*  maybe we need another segment (next extension) of descr. directory  */

   if (ext < totext)
      {
      ext ++ ;             /* prepare reading of next directory segment */
      diroff += fcbp->DIREXT;
      goto pos_loop;                    /* and loop more... */
      }

   *numbuf = --npos;		/* remove descriptor directory itself */
   
   return (ERR_NORMAL);
   }


/*  here for all the other functions */


nullo = -1;
(void) MID_RDSCRC(chanl,fcbp->PTRLDB,0,dirfirst,dirlen,dscdir,&nullo);
cpt = dscdir;

for (nr=0; nr<dirlen; nr+=fcbp->DIRENTRY)   /* take chunks of 30 chars. */
   {
   if ((*cpt != ' ') && (*(cpt+15) != 'H'))
      {
      if (++npos == snpos) 		/* always increment `npos' */
         {
         flag = 1;
         CGN_COPYALL(dscpntr,cpt,fcbp->DIRENTRY);
         goto descr_found;
         }
      }
   cpt += fcbp->DIRENTRY;
   }


/*  we need another segment (next extension) of descr. directory  */

if (ext < totext)
   {
   ext ++ ;             /* prepare reading of next directory segment */
   diroff += fcbp->DIREXT;
   goto pos_loop;                    /* and loop more... */
   }


descr_found:
if (snpos == 1)					/* process DIRECTORY.MIDAS */
   DSCDIR_ENTRY.NOELEM = (unsigned short int) dirused;	/* s.i => uns. s.i */
	
switch(fno)
   {
   case 1:
   MID_NAMINFO(flag,DSC_PNTR->NAME,buf,lbuf);
   break;

   case 2:
   MID_TYPINFO(flag,DSC_PNTR->TYPE,DSC_PNTR->BYTELEM,buf,lbuf);
   break;

   case 3:
   *numbuf = DSC_PNTR->NOELEM;
   break;

   case 4:
   MID_NAMINFO(flag,DSC_PNTR->NAME,buf,lbuf);
   if (lbuf > 15) 				/* old descr. length = 15 */
      {
      buf[15] = ',';
      lbuf -= 16;
      MID_TYPINFO(flag,DSC_PNTR->TYPE,DSC_PNTR->BYTELEM,buf+16,lbuf);
      }
   break;

   default:			/* case 5 */
   MID_NAMINFO(flag,DSC_PNTR->NAME,buf,lbuf);
   if (lbuf > 15) 
      {
      buf[15] = ',';
      lbuf -= 16;
      MID_TYPINFO(flag,DSC_PNTR->TYPE,DSC_PNTR->BYTELEM,buf+16,lbuf);
      }
   *numbuf++ = DSC_PNTR->NOELEM;
   *numbuf = DSC_PNTR->BYTELEM;
   }

return status;

	
end_of_it:
MID_E2(7,imno,"SCDINF: ",status,1);
return status;
}

/*

*/

ZSCDINF(chanl,imno,snpos,fno,buf,lbuf,numbuf)
/*++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE
provide info about descriptor at position 'npos'
.ALGORITHM
go through the descriptor directory and provide the desired info
.RETURNS
return always 0 (checks already done in SCDINF)
--------------------------------------------------*/

int  chanl;	/* IN: I/O chanl of frame */
int  imno	/* IN: no. of data frame  */;
int  snpos	/* IN: incremented position of descriptor */;
int  fno	/* IN: specify desired info, 
                        1 = NAME, 2 = TYPE, 3 = SIZE  
                        4 = NAME,TYPE separated by a comma in `buf'
                        5 = NAME,TYPE separated by a comma in `buf'
                            and SIZE,BYTELEM in `numbuf'
		       99 = return total no. of descriptors */;
char   *buf	/* IN: buffer for character descriptor info */;
int lbuf	/* IN: max. length of buffer above  */;
int *numbuf	/* IN: return buffer for numerical data  */;

{ 	
int   flag, ext;
int   dirfirst, dirused, dirsize, dirlen, diroff, totext;
int   nullo;
int   mm, dirbytes, npos;
register int  nr;

char  *dscpntr;
register char  *cpt;

struct FCT_STRUCT  *fctpntr;

struct FCB_STRUCT  *fcbp;

static int  mapdscdir = -1;
static char  *dscdir;



 
fctpntr = FCT.ENTRIES + imno;
fcbp = fctpntr->FZP;

dirused = fcbp->DFILLED;
dirsize = fcbp->DSIZE;
totext = dirsize/fcbp->DIREXT;      /* total no. of extensions for directory */
dscpntr = (char *) &ZDSCDIR_ENTRY;

ext = 1;
npos = 0;
diroff = 0;
flag = 0;


pos_loop:
dirfirst = diroff + 1;
dirlen = dirused - diroff;
if (fcbp->DIREXT < dirlen) dirlen = fcbp->DIREXT;    /* read max. 1500 chars. */


if (mapdscdir < 0)                      /* allocate working buffer */
   {                                    /* for descr. directory */
   dscdir = malloc((unsigned int)fcbp->DIREXT);            /* = 1500 */
   mapdscdir = 1;
   }


if (fno > 90)
   {
   nullo = -1;
   (void) MID_RDSCRC(chanl,fcbp->PTRLDB,0,dirfirst,dirlen,dscdir,&nullo);
   cpt = dscdir;

   nr = 0;
   while (nr < dirlen)
      {
      if (*cpt == '3')
         mm = 3;
      else if (*cpt == '2')
         mm = 2;
      else
         mm = 1;
      dirbytes = mm * fcbp->DIRENTRY;
      if (*(cpt+1) != '_') npos ++;	/* test if just fill_in block */

      cpt += dirbytes;  nr += dirbytes;
      }


   /*  maybe we need another segment (next extension) of descr. directory  */

   if (ext < totext)
      {
      ext ++ ;             /* prepare reading of next directory segment */
      diroff += fcbp->DIREXT;
      goto pos_loop;                    /* and loop more... */
      }

   *numbuf = --npos;		/* remove descriptor directory itself */
   return (ERR_NORMAL);
   }


/*  here for all the other functions */

nullo = -1;
(void) MID_RDSCRC(chanl,fcbp->PTRLDB,0,dirfirst,dirlen,dscdir,&nullo);
cpt = dscdir;

nr = 0;
while (nr < dirlen)
   {
   if (*cpt == '3')
      mm = 3;
   else if (*cpt == '2')
      mm = 2;
   else
      mm = 1;
   dirbytes = mm * fcbp->DIRENTRY;

   if (*(cpt+1) != '_')              /* test if just fill_in block */
      {
      if (++npos == snpos) 		/* always increment `npos' */
         {
         CGN_COPYALL(dscpntr,cpt,dirbytes);
         flag = 1;
         goto descr_found;
         }
      }

   cpt += dirbytes;  nr += dirbytes;
   }


/*  we need another segment (next extension) of descr. directory  */

if (ext < totext)
   {
   ext ++ ;             /* prepare reading of next directory segment */
   diroff += fcbp->DIREXT;
   goto pos_loop;                    /* and loop more... */
   }


descr_found:
switch(fno)
   {
   case 1:
   if (flag == 0)			/* npos too large */
      CGN_FILL(buf,' ',lbuf);
   else
      {
      mm = (int) strlen(ZDSC_PNTR->NAME);
      if (mm > lbuf)
         (void) strncpy(buf,ZDSC_PNTR->NAME,lbuf);
      else
         (void) strcpy(buf,ZDSC_PNTR->NAME);
      }
   break;

   case 2:
   (void) MID_TYPINFO(flag,ZDSC_PNTR->TYPE,ZDSC_PNTR->BYTELEM,buf,lbuf);
   break;

   case 3:
   if (flag == 0)                       /* npos too large */
      *numbuf = 0;
   else
      {
      if (snpos == 1)
         *numbuf = dirused;
      else
         *numbuf = ZDSC_PNTR->NOELEM;
      }
   break;

   case 4:
   if (flag == 0)                       /* npos too large */
      CGN_FILL(buf,' ',lbuf);
   else
      {
      mm = (int) strlen(ZDSC_PNTR->NAME);
      if (mm > lbuf)
         (void) strncpy(buf,ZDSC_PNTR->NAME,lbuf);
      else
         {
         (void) strcpy(buf,ZDSC_PNTR->NAME);
         lbuf -= mm;
         if (lbuf > 1)
            {
            buf[mm] = ',';
            lbuf --;
            (void) MID_TYPINFO(flag,ZDSC_PNTR->TYPE,ZDSC_PNTR->BYTELEM,
                               buf+mm+1,lbuf);
            }
         }
      }
   break;

   default:				/* case 5 */
   if (flag == 0)                       /* npos too large */
      CGN_FILL(buf,' ',lbuf);
   else
      {
      mm = (int) strlen(ZDSC_PNTR->NAME);
      if (mm > lbuf)
         (void) strncpy(buf,ZDSC_PNTR->NAME,lbuf);
      else
         {
         (void) strcpy(buf,ZDSC_PNTR->NAME);
         lbuf -= mm;
         if (lbuf > 1)
            {
            buf[mm] = ',';
            lbuf --;
            (void) MID_TYPINFO(flag,ZDSC_PNTR->TYPE,ZDSC_PNTR->BYTELEM,
                               buf+mm+1,lbuf);
            }
      
         if (snpos == 1)
            *numbuf++ = dirused;
         else
            *numbuf++ = ZDSC_PNTR->NOELEM;
         *numbuf = ZDSC_PNTR->BYTELEM;
         }
      }
   }

return ERR_NORMAL;
}
