/* @(#)ihaprdm.c	15.1.1.1 (ES0-DMD) 01/12/01 16:32:41 */
/*===========================================================================
  Copyright (C) 1995 European Southern Observatory (ESO)
 
  This program is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public 
  License along with this program; if not, write to the Free 
  Software Foundation, Inc., 675 Massachusetss Ave, Cambridge, 
  MA 02139, USA.
 
  Corresponding concerning ESO-MIDAS should be addressed as follows:
	Internet e-mail: midas@eso.org
	Postal address: European Southern Observatory
			Data Management Division 
			Karl-Schwarzschild-Strasse 2
			D 85748 Garching bei Muenchen 
			GERMANY
===========================================================================*/

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.COPYRIGHT  (c)  1994   European Southern Observatory
.IDENT      ihaprdm.c
.LAUGUAGE   C
.AUTHOR     P.Grosbol   ESO/IPG
.KEYWORDS   IHAP data matrix, decode
.COMMENT    decode IHAP data martix with format 1,3 - image type
.VERSION    1.0  1988-Nov-06 : Creation,   PJG 
.VERSION    1.1  1990-Feb-04 : Change calling seq. of cvXX,   PJG 
.VERSION    1.2  1990-Feb-16 : Correct interlace,   PJG 
.VERSION    2.0  1991-Feb-17 : Change structure and call-seq.,   PJG 
.VERSION    2.1  1992-May-18 : Correct index error in min/max,   PJG 
.VERSION    2.2  1993-Oct-28 : Update to new SC + prototypes,   PJG 
.VERSION    2.3  1994-Apr-20 : Initiate pointer for I*2 formaterror,  PJG 
---------------------------------------------------------------------*/
#include   <computer.h>
#include   <fitsfmt.h>
#include   <fitsdef.h>
#include   <midas_def.h>

#define  BSIZE             16384   /* standard byte size for read    */
#define  FSIZE               512   /* I2->R4 conversion buffer size  */

#ifdef __STDC__
int ihaprdm(int mfd,BFDEF *bfdef,int size,char fmt)
#else
int ihaprdm(mfd,bfdef,size,fmt)
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.PURPOSE       read IHAP data matrix
.RETURN        error code: 0: OK, -1: error
---------------------------------------------------------------------*/
int      mfd;                  /* IN: MIDAS file desc.               */
BFDEF    *bfdef;               /* IN: FITS/IHAP header parm's        */
int      size;                 /* IN: size of data matrix (bytes)    */
char     fmt;                  /* IN: MIDAS file format, No/Orig/Fp  */
#endif
{
  register short  *pss, *pse, s, mnvs, mxvs;
  char            *pc;
  short           *psb;
  int             np, i, no, ny, nb, unit[4];
  int             felem;
  float           *pfb, mnvf, mxvf, cuts[2], fcvt[FSIZE];

  felem = 1;
  mnvf = MAXFLOAT; mxvf = -MAXFLOAT;
  mnvs = MAXSHORT; mxvs = MINSHORT;
  if (bfdef->sflag) {             /* process data for interlace mode */
     if (bfdef->bitpix == -32) {  /* check data format - only int.   */
        SCTPUT("Error: only integer format can be interlaced!");
        SCFCLO(mfd); mfd = -1; return -1;
     }
     if (BSIZE<bfdef->data[0].naxis) {     /* check length of line   */
        SCTPUT("Error: X-axis too long to be interlaced!");
        SCFCLO(mfd); mfd = -1; return -1;
     }

     nb = 2 * bfdef->data[0].naxis;     /* no. of bytes per line     */
     for (ny=0; ny<bfdef->data[1].naxis; ny++) {    /* line by line  */
         if ((no=dread(&pc,nb))<nb) {
            SCTPUT("Error: reading data matrix");
            SCFCLO(mfd); mfd = -1; return -1;
         }
         size -= no;
	 psb = (short *) pc;
         cvi2(psb,bfdef->data[0].naxis,0);
         if (ny&1) {                    /* interlace odd line        */
            pss = psb; pse = pss + bfdef->data[0].naxis - 1;
            i = bfdef->data[0].naxis >> 1;
            while (i--) { s = *pss; *pss++ = *pse; *pse-- = s; }
         }

         if (fmt=='O') {
            SCFPUT(mfd,felem,bfdef->data[0].naxis,(char *) psb);
	    felem += bfdef->data[0].naxis;
         }
         else {
            no = bfdef->data[0].naxis;
            while (0<no) {
               np = (no<FSIZE) ? no : FSIZE;
	       pss = psb;
               for (i=0; i<np; i++) fcvt[i] = *pss++;
               SCFPUT(mfd,felem,np,(char *) fcvt);
               felem += np; no -= np;
            }
         }
	 i = bfdef->data[0].naxis; pss = psb;
	 while (i--) {
	   if (*pss<mnvs) mnvs = *pss;
	   else if (mxvs<*pss) mxvs = *pss;
	   pss++;
	 }
     }
  }
  else {                                   /* process normal data       */
     while (0<size) {                      /* read file in big blocks   */
        nb = (size<BSIZE) ? size : BSIZE;

        if ((nb=dread(&pc,nb))<0) {
           SCTPUT("Error: reading data matrix");
           SCFCLO(mfd); mfd = -1; return -1;
        }
        size -= nb;
	i = (bfdef->bitpix == 16) ? 2 : 4;
        no = nb / i;
        if (bfdef->bitpix == -32) {          /* convert fp-format data    */
           pfb = (float *) pc;
           cvr4(pfb,no,0);
           SCFPUT(mfd,felem,no,(char *) pfb); felem += no;
	   i = no;
	   while (i--) {
	     if (*pfb<mnvf) mnvf = *pfb;
	     else if (mxvf<*pfb) mxvf = *pfb;
	     pfb++;
	   }
        }
        else {
           psb = (short *) pc;
           cvi2(psb,no,0);
           if (fmt=='O') {
              SCFPUT(mfd,felem,no,(char *) psb); felem += no;
           }
           else
              while (0<no) {
                 nb = (no<FSIZE) ? no : FSIZE;
		 pss = psb;
                 for (i=0; i<nb; i++) fcvt[i] = *pss++;
                 SCFPUT(mfd,felem,nb,(char *) fcvt);
                 felem += nb; no -= nb;
              }
	   i = no; pss = psb;
	   while (i--) {
	     if (*pss<mnvs) mnvs = *pss;
	     else if (mxvs<*pss) mxvs = *pss;
	     pss++;
	   }
        }
     }
  }
  if (bfdef->bitpix == 16) { cuts[0] = mnvs; cuts[1] = mxvs; }
   else { cuts[0] = mnvf; cuts[1] = mxvf; }
  SCDWRR(mfd,"LHCUTS",cuts,3,2,unit);

  if (0 <= mfd) SCFCLO(mfd); mfd = -1;
  return 0;
}
